from typing import Union

from google.cloud.aiplatform.base import VertexAiResourceNoun
from google.cloud.aiplatform.jobs import _RunnableJob
from google.cloud.aiplatform.training_jobs import _TrainingJob
from google.cloud.aiplatform_v1.types import JobState, PipelineState

from datahub.metadata.schema_classes import RunResultTypeClass


def get_automl_job_result_type(state: PipelineState) -> Union[str, RunResultTypeClass]:
    state_mapping = {
        PipelineState.PIPELINE_STATE_SUCCEEDED: RunResultTypeClass.SUCCESS,
        PipelineState.PIPELINE_STATE_FAILED: RunResultTypeClass.FAILURE,
        PipelineState.PIPELINE_STATE_CANCELLED: "Cancelled",
        PipelineState.PIPELINE_STATE_PAUSED: "Paused",
        PipelineState.PIPELINE_STATE_QUEUED: "Queued",
        PipelineState.PIPELINE_STATE_RUNNING: "Running",
        PipelineState.PIPELINE_STATE_UNSPECIFIED: "Unspecific",
    }

    return state_mapping.get(state, "UNKNOWN")


def get_custom_job_result_type(state: JobState) -> Union[str, RunResultTypeClass]:
    state_mapping = {
        JobState.JOB_STATE_SUCCEEDED: RunResultTypeClass.SUCCESS,
        JobState.JOB_STATE_FAILED: RunResultTypeClass.FAILURE,
        JobState.JOB_STATE_CANCELLED: "Cancelled",
        JobState.JOB_STATE_PAUSED: "Paused",
        JobState.JOB_STATE_QUEUED: "Queued",
        JobState.JOB_STATE_RUNNING: "Running",
        JobState.JOB_STATE_CANCELLING: "Cancelling",
        JobState.JOB_STATE_EXPIRED: "Expired",
        JobState.JOB_STATE_UPDATING: "Updating",
    }
    return state_mapping.get(state, "UNKNOWN")


def get_job_result_status(job: VertexAiResourceNoun) -> Union[str, RunResultTypeClass]:
    if isinstance(job, _TrainingJob) and isinstance(job.state, PipelineState):
        return get_automl_job_result_type(job.state)
    elif isinstance(job, _RunnableJob) and isinstance(job.state, JobState):
        return get_custom_job_result_type(job.state)
    return "UNKNOWN"


def get_execution_result_status(status: int) -> Union[str, RunResultTypeClass]:
    """
    State of the execution.
    STATE_UNSPECIFIED = 0
    PENDING = 1
    RUNNING = 2
    SUCCEEDED = 3
    FAILED = 4
    """
    status_mapping = {
        0: "STATE_UNSPECIFIED",
        1: "PENDING",
        2: "RUNNING",
        3: RunResultTypeClass.SUCCESS,
        4: RunResultTypeClass.FAILURE,
    }
    return status_mapping.get(status, "UNKNOWN")


def is_status_for_run_event_class(status: Union[str, RunResultTypeClass]) -> bool:
    return status in [RunResultTypeClass.SUCCESS, RunResultTypeClass.FAILURE]



# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import DynamicFormAssignmentClass
from .....schema_classes import FormActorAssignmentClass
from .....schema_classes import FormInfoClass
from .....schema_classes import FormPromptClass
from .....schema_classes import FormPromptTypeClass
from .....schema_classes import FormTypeClass
from .....schema_classes import StructuredPropertyParamsClass


DynamicFormAssignment = DynamicFormAssignmentClass
FormActorAssignment = FormActorAssignmentClass
FormInfo = FormInfoClass
FormPrompt = FormPromptClass
FormPromptType = FormPromptTypeClass
FormType = FormTypeClass
StructuredPropertyParams = StructuredPropertyParamsClass

# fmt: on



import logging
import time
from typing import Dict, Iterable, List, Optional, Union

from datahub.configuration.common import ConfigModel, DynamicTypedConfig
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.common import PipelineContext, RecordEnvelope
from datahub.ingestion.api.sink import NoopWriteCallback, Sink
from datahub.ingestion.sink.sink_registry import sink_registry
from datahub.lite.lite_local import (
    Browseable,
    DataHubLiteLocal,
    Searchable,
    SearchFlavor,
)
from datahub.lite.lite_registry import lite_registry
from datahub.metadata.schema_classes import MetadataChangeEventClass, _Aspect

logger = logging.getLogger(__name__)


class LiteLocalConfig(ConfigModel):
    type: str
    config: dict = {}
    forward_to: Optional[DynamicTypedConfig] = None


class DataHubLiteWrapper(DataHubLiteLocal):
    def __init__(self, lite_impl: DataHubLiteLocal, forward_to: Sink):
        self.lite = lite_impl
        self.forward_to = forward_to

    def write(
        self,
        record: Union[
            MetadataChangeEventClass,
            MetadataChangeProposalWrapper,
        ],
    ) -> None:
        self.lite.write(record)
        record_envelope = RecordEnvelope(record=record, metadata={})
        self.forward_to.write_record_async(
            record_envelope=record_envelope, write_callback=NoopWriteCallback()
        )

    def close(self) -> None:
        self.lite.close()
        self.forward_to.close()

    def location(self) -> str:
        return self.lite.location()

    def destroy(self) -> None:
        return self.lite.destroy()

    def list_ids(self) -> Iterable[str]:
        yield from self.lite.list_ids()

    def get(
        self,
        id: str,
        aspects: Optional[List[str]],
        typed: bool = False,
        as_of: Optional[int] = None,
        details: Optional[bool] = False,
    ) -> Optional[Dict[str, Union[str, dict, _Aspect]]]:
        return self.get(id, aspects, typed, as_of, details)

    def search(
        self,
        query: str,
        flavor: SearchFlavor,
        aspects: Optional[List[str]] = None,
        snippet: bool = True,
    ) -> Iterable[Searchable]:
        aspects = aspects or []
        yield from self.lite.search(query, flavor, aspects, snippet)

    def ls(self, path: str) -> List[Browseable]:
        return self.lite.ls(path)

    def get_all_entities(
        self, typed: bool = False
    ) -> Iterable[Dict[str, Union[dict, _Aspect]]]:
        yield from self.lite.get_all_entities(typed)

    def get_all_aspects(self) -> Iterable[MetadataChangeProposalWrapper]:
        yield from self.lite.get_all_aspects()

    def reindex(self) -> None:
        self.lite.reindex()


def get_datahub_lite(config_dict: dict, read_only: bool = False) -> "DataHubLiteLocal":
    lite_local_config = LiteLocalConfig.parse_obj(config_dict)

    lite_type = lite_local_config.type
    try:
        lite_class = lite_registry.get(lite_type)
    except KeyError as e:
        raise Exception(
            f"Failed to find a registered lite implementation for {lite_type}. Valid values are {[k for k in lite_registry.mapping.keys()]}"
        ) from e

    lite_specific_config = lite_class.get_config_class().parse_obj(
        lite_local_config.config
    )
    lite = lite_class(lite_specific_config)
    # we only set up forwarding if forwarding config is present and read_only is set to False
    if lite_local_config.forward_to and not read_only:
        forward_sink_class = sink_registry.get(lite_local_config.forward_to.type)
        if forward_sink_class is not None:
            current_time = int(time.time() * 1000.0)
            try:
                forward_to = forward_sink_class.create(
                    lite_local_config.forward_to.config or {},
                    PipelineContext(run_id=f"lite-forward_{current_time}"),
                )
                return DataHubLiteWrapper(lite, forward_to)
            except Exception as e:
                logger.warning(
                    f"Failed to set up forwarding due to {e}, will not forward events"
                )
                logger.debug(
                    "Failed to set up forwarding, will not forward events", exc_info=e
                )
                return lite
        else:
            raise Exception(
                f"Failed to find a registered forwarding sink for type {lite_local_config.forward_to.type}. Valid values are {[k for k in sink_registry.mapping.keys()]}"
            )
    else:
        return lite



import json
import logging
from dataclasses import dataclass
from typing import Dict, List, Optional, Sequence, Union

from requests import Response

from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.metadata.com.linkedin.pegasus2avro.mxe import (
    MetadataChangeProposal,
)

logger = logging.getLogger(__name__)


@dataclass
class TraceData:
    trace_id: str
    data: Dict[str, List[str]]

    def __post_init__(self) -> None:
        if not self.trace_id:
            raise ValueError("trace_id cannot be empty")
        if not isinstance(self.data, dict):
            raise TypeError("data must be a dictionary")


def _extract_trace_id(
    response: Response, trace_header: str = "traceparent"
) -> Optional[str]:
    """
    Extract trace ID from response headers.
    Args:
        response: HTTP response object
        trace_header: Name of the trace header to use
    Returns:
        Trace ID if found and response is valid, None otherwise
    """
    if not 200 <= response.status_code < 300:
        logger.debug(f"Invalid status code: {response.status_code}")
        return None

    trace_id = response.headers.get(trace_header)
    if not trace_id:
        logger.debug(f"Missing trace header: {trace_header}")
        return None

    return trace_id


def extract_trace_data(
    response: Response,
    aspects_to_trace: Optional[List[str]] = None,
    trace_header: str = "traceparent",
) -> Optional[TraceData]:
    """
    Extract trace data from a response object.
    Args:
        response: HTTP response object
        aspects_to_trace: Optional list of aspect names to extract. If None, extracts all aspects.
        trace_header: Name of the trace header to use (default: "traceparent")
    Returns:
        TraceData object if successful, None otherwise
    Raises:
        JSONDecodeError: If response body cannot be decoded as JSON
    """
    trace_id = _extract_trace_id(response, trace_header)
    if not trace_id:
        return None

    try:
        json_data = response.json()
        if not isinstance(json_data, list):
            logger.debug("JSON data is not a list")
            return None

        data: Dict[str, List[str]] = {}

        for item in json_data:
            urn = item.get("urn")
            if not urn:
                logger.debug(f"Skipping item without URN: {item}")
                continue

            if aspects_to_trace is None:
                aspect_names = [
                    k for k, v in item.items() if k != "urn" and v is not None
                ]
            else:
                aspect_names = [
                    field for field in aspects_to_trace if item.get(field) is not None
                ]

            data[urn] = aspect_names

        return TraceData(trace_id=trace_id, data=data)

    except json.JSONDecodeError as e:
        logger.error(f"Failed to decode JSON response: {e}")
        return None


def extract_trace_data_from_mcps(
    response: Response,
    mcps: Sequence[Union[MetadataChangeProposal, MetadataChangeProposalWrapper]],
    aspects_to_trace: Optional[List[str]] = None,
    trace_header: str = "traceparent",
) -> Optional[TraceData]:
    """
    Extract trace data from a response object and populate data from provided MCPs.
    Args:
        response: HTTP response object used only for trace_id extraction
        mcps: List of MCP URN and aspect data
        aspects_to_trace: Optional list of aspect names to extract. If None, extracts all aspects.
        trace_header: Name of the trace header to use (default: "traceparent")
    Returns:
        TraceData object if successful, None otherwise
    """
    trace_id = _extract_trace_id(response, trace_header)
    if not trace_id:
        return None

    data: Dict[str, List[str]] = {}
    try:
        for mcp in mcps:
            entity_urn = getattr(mcp, "entityUrn", None)
            aspect_name = getattr(mcp, "aspectName", None)

            if not entity_urn or not aspect_name:
                logger.debug(f"Skipping MCP with missing URN or aspect name: {mcp}")
                continue

            if aspects_to_trace is not None and aspect_name not in aspects_to_trace:
                continue

            if entity_urn not in data:
                data[entity_urn] = []

            data[entity_urn].append(aspect_name)

        return TraceData(trace_id=trace_id, data=data)

    except AttributeError as e:
        logger.error(f"Error processing MCPs: {e}")
        return None



import functools
from typing import Iterable

from datahub.emitter.mce_builder import get_sys_time
from datahub.ingestion.api.common import PipelineContext, RecordEnvelope
from datahub.ingestion.api.transform import Transformer
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.transformer.auto_helper_transformer import AutoHelperTransformer
from datahub.metadata.schema_classes import SystemMetadataClass


def auto_system_metadata(
    ctx: PipelineContext,
    stream: Iterable[MetadataWorkUnit],
) -> Iterable[MetadataWorkUnit]:
    if not ctx.pipeline_config:
        raise ValueError("Pipeline config is required for system metadata")
    set_system_metadata = ctx.pipeline_config.flags.set_system_metadata
    set_pipeline_name = ctx.pipeline_config.flags.set_system_metadata_pipeline_name

    for workunit in stream:
        if set_system_metadata:
            workunit.metadata.systemMetadata = SystemMetadataClass(
                lastObserved=get_sys_time(), runId=ctx.run_id
            )
            if set_pipeline_name:
                workunit.metadata.systemMetadata.pipelineName = ctx.pipeline_name

        yield workunit


class SystemMetadataTransformer(Transformer):
    def __init__(self, ctx: PipelineContext):
        self._inner_transformer = AutoHelperTransformer(
            functools.partial(auto_system_metadata, ctx)
        )

    def transform(
        self, record_envelopes: Iterable[RecordEnvelope]
    ) -> Iterable[RecordEnvelope]:
        yield from self._inner_transformer.transform(record_envelopes)

    @classmethod
    def create(cls, config_dict: dict, ctx: PipelineContext) -> Transformer:
        raise NotImplementedError(f"{cls.__name__} cannot be created from config")



import logging
from collections import defaultdict
from typing import Any, Dict, List, Union

import sqlalchemy as sa
from sqlalchemy.engine import Connection, Engine
from sqlalchemy.engine.reflection import Inspector

from datahub.ingestion.source.common.data_reader import DataReader
from datahub.utilities.perf_timer import PerfTimer

logger: logging.Logger = logging.getLogger(__name__)


class SqlAlchemyTableDataReader(DataReader):
    @staticmethod
    def create(inspector: Inspector) -> "SqlAlchemyTableDataReader":
        return SqlAlchemyTableDataReader(conn=inspector.bind)

    def __init__(
        self,
        conn: Union[Engine, Connection],
    ) -> None:
        self.connection = conn.engine.connect()

    def _table(self, table_id: List[str]) -> sa.Table:
        return sa.Table(
            table_id[-1],
            sa.MetaData(),
            schema=table_id[-2] if len(table_id) > 1 else None,
        )

    def get_sample_data_for_table(
        self, table_id: List[str], sample_size: int, **kwargs: Any
    ) -> Dict[str, list]:
        """
        For sqlalchemy, table_id should be in form (schema_name, table_name)
        """

        logger.debug(f"Collecting sample values for table {'.'.join(table_id)}")

        with PerfTimer() as timer:
            column_values: Dict[str, list] = defaultdict(list)
            table = self._table(table_id)

            query: Any

            # limit doesn't compile properly for oracle so we will append rownum to query string later
            if self.connection.dialect.name.lower() == "oracle":
                raw_query = sa.select([sa.text("*")]).select_from(table)

                query = str(
                    raw_query.compile(
                        self.connection, compile_kwargs={"literal_binds": True}
                    )
                )
                query += "\nAND ROWNUM <= %d" % sample_size
            else:
                query = sa.select([sa.text("*")]).select_from(table).limit(sample_size)
            query_results = self.connection.execute(query)

            # Not ideal - creates a parallel structure in column_values. Can we use pandas here ?
            for row in query_results.fetchall():
                for col, col_value in row._mapping.items():
                    column_values[col].append(col_value)
            time_taken = timer.elapsed_seconds()
            logger.debug(
                f"Finished collecting sample values for table {'.'.join(table_id)};"
                f"took {time_taken:.3f} seconds"
            )
        return column_values

    def close(self) -> None:
        self.connection.close()



from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional


@dataclass
class Workspace:
    name: str


@dataclass
class Status:
    name: str


@dataclass
class Category:
    name: str
    description: Optional[str] = None


@dataclass
class Collection:
    name: str


@dataclass(frozen=True)
class Owner:
    email: str


@dataclass
class Analytics:
    appviews_all_time: Optional[int]
    appviews_last_7_days: Optional[int]
    appviews_last_14_days: Optional[int]
    appviews_last_30_days: Optional[int]
    last_viewed_at: Optional[datetime]


@dataclass
class Project:
    id: str
    title: str
    description: Optional[str]
    last_edited_at: Optional[datetime] = None
    created_at: Optional[datetime] = None
    status: Optional[Status] = None
    categories: Optional[List[Category]] = None  # TODO: emit category description!
    collections: Optional[List[Collection]] = None
    creator: Optional[Owner] = None
    owner: Optional[Owner] = None
    analytics: Optional[Analytics] = None


@dataclass
class Component:
    id: str
    title: str
    description: Optional[str]
    last_edited_at: Optional[datetime] = None
    created_at: Optional[datetime] = None
    status: Optional[Status] = None
    categories: Optional[List[Category]] = None
    collections: Optional[List[Collection]] = None
    creator: Optional[Owner] = None
    owner: Optional[Owner] = None
    analytics: Optional[Analytics] = None



from enum import Enum
from typing import List, Optional

from pydantic import Field, SecretStr

from datahub.configuration import ConfigModel


class OAuthIdentityProvider(Enum):
    MICROSOFT = "microsoft"
    OKTA = "okta"


class OAuthConfiguration(ConfigModel):
    provider: OAuthIdentityProvider = Field(
        description="Identity provider for oauth."
        "Supported providers are microsoft and okta."
    )
    authority_url: str = Field(description="Authority url of your identity provider")
    client_id: str = Field(description="client id of your registered application")
    scopes: List[str] = Field(description="scopes required to connect to snowflake")
    use_certificate: bool = Field(
        description="Do you want to use certificate and private key to authenticate using oauth",
        default=False,
    )
    client_secret: Optional[SecretStr] = Field(
        None, description="client secret of the application if use_certificate = false"
    )
    encoded_oauth_public_key: Optional[str] = Field(
        None, description="base64 encoded certificate content if use_certificate = true"
    )
    encoded_oauth_private_key: Optional[str] = Field(
        None, description="base64 encoded private key content if use_certificate = true"
    )



import datetime
import logging
import time
from typing import Any, Dict, Iterator, Optional

from pydantic import BaseModel, Field

from datahub.configuration.common import ConfigModel
from datahub.ingestion.api.source import SourceReport
from datahub.ingestion.graph.client import DataHubGraph

logger = logging.getLogger(__name__)

DATAHUB_EXECUTION_REQUEST_ENTITY_NAME = "dataHubExecutionRequest"
DATAHUB_EXECUTION_REQUEST_KEY_ASPECT_NAME = "dataHubExecutionRequestKey"
DATAHUB_EXECUTION_REQUEST_INPUT_ASPECT_NAME = "dataHubExecutionRequestInput"
DATAHUB_EXECUTION_REQUEST_RESULT_ASPECT_NAME = "dataHubExecutionRequestResult"


class DatahubExecutionRequestCleanupConfig(ConfigModel):
    keep_history_min_count: int = Field(
        10,
        description="Minimum number of execution requests to keep, per ingestion source",
    )

    keep_history_max_count: int = Field(
        1000,
        description="Maximum number of execution requests to keep, per ingestion source",
    )

    keep_history_max_days: int = Field(
        90,
        description="Maximum number of days to keep execution requests for, per ingestion source",
    )

    batch_read_size: int = Field(
        100,
        description="Number of records per read operation",
    )

    enabled: bool = Field(
        True,
        description="Global switch for this cleanup task",
    )

    runtime_limit_seconds: int = Field(
        default=3600,
        description="Maximum runtime in seconds for the cleanup task",
    )

    limit_entities_delete: Optional[int] = Field(
        10000, description="Max number of execution requests to hard delete."
    )

    max_read_errors: int = Field(
        default=10,
        description="Maximum number of read errors before aborting",
    )

    def keep_history_max_milliseconds(self):
        return self.keep_history_max_days * 24 * 3600 * 1000


class DatahubExecutionRequestCleanupReport(SourceReport):
    ergc_records_read: int = 0
    ergc_records_preserved: int = 0
    ergc_records_deleted: int = 0
    ergc_read_errors: int = 0
    ergc_delete_errors: int = 0
    ergc_start_time: Optional[datetime.datetime] = None
    ergc_end_time: Optional[datetime.datetime] = None
    ergc_delete_limit_reached: bool = False
    ergc_runtime_limit_reached: bool = False


class CleanupRecord(BaseModel):
    urn: str
    request_id: str
    status: str
    ingestion_source: str
    requested_at: int


class DatahubExecutionRequestCleanup:
    def __init__(
        self,
        graph: DataHubGraph,
        report: DatahubExecutionRequestCleanupReport,
        config: Optional[DatahubExecutionRequestCleanupConfig] = None,
    ) -> None:
        self.graph = graph
        self.report = report
        self.instance_id = int(time.time())
        self.last_print_time = 0.0

        if config is not None:
            self.config = config
        else:
            self.config = DatahubExecutionRequestCleanupConfig()

    def _print_report(self) -> None:
        time_taken = round(time.time() - self.last_print_time, 1)
        # Print report every 2 minutes
        if time_taken > 120:
            self.last_print_time = time.time()
            logger.info(f"\n{self.report.as_string()}")

    def _to_cleanup_record(self, entry: Dict) -> CleanupRecord:
        input_aspect = (
            entry.get("aspects", {})
            .get(DATAHUB_EXECUTION_REQUEST_INPUT_ASPECT_NAME, {})
            .get("value", {})
        )
        result_aspect = (
            entry.get("aspects", {})
            .get(DATAHUB_EXECUTION_REQUEST_RESULT_ASPECT_NAME, {})
            .get("value", {})
        )
        key_aspect = (
            entry.get("aspects", {})
            .get(DATAHUB_EXECUTION_REQUEST_KEY_ASPECT_NAME, {})
            .get("value", {})
        )
        return CleanupRecord(
            urn=entry.get("urn"),
            request_id=key_aspect.get("id"),
            requested_at=input_aspect.get("requestedAt", 0),
            status=result_aspect.get("status", "PENDING"),
            ingestion_source=input_aspect.get("source", {}).get("ingestionSource", ""),
        )

    def _scroll_execution_requests(
        self, overrides: Optional[Dict[str, Any]] = None
    ) -> Iterator[CleanupRecord]:
        overrides = overrides or {}
        headers: Dict[str, Any] = {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
        params = {
            "aspectNames": [
                DATAHUB_EXECUTION_REQUEST_KEY_ASPECT_NAME,
                DATAHUB_EXECUTION_REQUEST_INPUT_ASPECT_NAME,
                DATAHUB_EXECUTION_REQUEST_RESULT_ASPECT_NAME,
            ],
            "count": str(self.config.batch_read_size),
            "sort": "requestTimeMs",
            "sortOrder": "DESCENDING",
            "systemMetadata": "false",
            "skipCache": "true",
        }
        params.update(overrides)

        while True:
            if self._reached_runtime_limit():
                break
            if self.report.ergc_read_errors >= self.config.max_read_errors:
                self.report.failure(
                    title="Too many read errors, aborting",
                    message="Too many read errors, aborting",
                    context=str(self.instance_id),
                )
                break
            try:
                url = f"{self.graph.config.server}/openapi/v2/entity/{DATAHUB_EXECUTION_REQUEST_ENTITY_NAME}"
                response = self.graph._session.get(url, headers=headers, params=params)
                response.raise_for_status()
                document = response.json()

                entries = document.get("results", [])
                for entry in entries:
                    yield self._to_cleanup_record(entry)

                if "scrollId" not in document:
                    break
                params["scrollId"] = document["scrollId"]
            except Exception as e:
                self.report.failure(
                    title="Failed to fetch next batch of execution requests",
                    message="Failed to fetch next batch of execution requests",
                    context=str(self.instance_id),
                    exc=e,
                )
                self.report.ergc_read_errors += 1

    def _scroll_garbage_records(self):
        state: Dict[str, Dict] = {}

        now_ms = int(time.time()) * 1000
        running_guard_timeout = now_ms - 30 * 24 * 3600 * 1000

        for entry in self._scroll_execution_requests():
            self._print_report()
            self.report.ergc_records_read += 1
            key = entry.ingestion_source

            # Always delete corrupted records
            if not key:
                logger.warning(
                    f"ergc({self.instance_id}): will delete corrupted entry with missing source key: {entry}"
                )
                yield entry
                continue

            if key not in state:
                state[key] = {}
                state[key]["cutoffTimestamp"] = (
                    entry.requested_at - self.config.keep_history_max_milliseconds()
                )

            state[key]["count"] = state[key].get("count", 0) + 1

            # Do not delete if number of requests is below minimum
            if state[key]["count"] < self.config.keep_history_min_count:
                self.report.ergc_records_preserved += 1
                continue

            # Do not delete if number of requests do not exceed allowed maximum,
            # or the cutoff date.
            if (state[key]["count"] < self.config.keep_history_max_count) and (
                entry.requested_at > state[key]["cutoffTimestamp"]
            ):
                self.report.ergc_records_preserved += 1
                continue

            # Do not delete if status is RUNNING or PENDING and created within last month. If the record is >month old and it did not
            # transition to a final state within that timeframe, it likely has no value.
            if entry.requested_at > running_guard_timeout and entry.status in [
                "RUNNING",
                "PENDING",
            ]:
                self.report.ergc_records_preserved += 1
                continue

            # Otherwise delete current record
            logger.info(
                (
                    f"ergc({self.instance_id}): going to delete {entry.request_id} in source {key}; "
                    f"source count: {state[key]['count']}; "
                    f"source cutoff: {state[key]['cutoffTimestamp']}; "
                    f"record timestamp: {entry.requested_at}."
                )
            )
            yield entry

    def _delete_entry(self, entry: CleanupRecord) -> None:
        try:
            self.graph.delete_entity(entry.urn, True)
            self.report.ergc_records_deleted += 1
        except Exception as e:
            self.report.ergc_delete_errors += 1
            self.report.failure(
                title="Failed to delete ExecutionRequest",
                message="Failed to delete ExecutionRequest",
                context=str(self.instance_id),
                exc=e,
            )

    def _reached_runtime_limit(self) -> bool:
        if (
            self.config.runtime_limit_seconds
            and self.report.ergc_start_time
            and (
                datetime.datetime.now() - self.report.ergc_start_time
                >= datetime.timedelta(seconds=self.config.runtime_limit_seconds)
            )
        ):
            self.report.ergc_runtime_limit_reached = True
            logger.info(f"ergc({self.instance_id}): max runtime reached.")
            return True
        return False

    def _reached_delete_limit(self) -> bool:
        if (
            self.config.limit_entities_delete
            and self.report.ergc_records_deleted >= self.config.limit_entities_delete
        ):
            logger.info(
                f"ergc({self.instance_id}): max delete limit reached: {self.config.limit_entities_delete}."
            )
            self.report.ergc_delete_limit_reached = True
            return True
        return False

    def run(self) -> None:
        if not self.config.enabled:
            logger.info(
                f"ergc({self.instance_id}): ExecutionRequest cleaner is disabled."
            )
            return
        self.report.ergc_start_time = datetime.datetime.now()

        logger.info(
            (
                f"ergc({self.instance_id}): Starting cleanup of ExecutionRequest records; "
                f"max days: {self.config.keep_history_max_days}, "
                f"min records: {self.config.keep_history_min_count}, "
                f"max records: {self.config.keep_history_max_count}."
            )
        )

        for entry in self._scroll_garbage_records():
            if self._reached_runtime_limit() or self._reached_delete_limit():
                break
            self._delete_entry(entry)

        self.report.ergc_end_time = datetime.datetime.now()
        logger.info(
            f"ergc({self.instance_id}): Finished cleanup of ExecutionRequest records."
        )



from typing import Union

from typing_extensions import Self

from datahub.emitter.mcp_patch_builder import MetadataPatchProposal
from datahub.metadata.schema_classes import (
    GlossaryTermAssociationClass as Term,
    GlossaryTermsClass,
)
from datahub.metadata.urns import GlossaryTermUrn, Urn


class HasTermsPatch(MetadataPatchProposal):
    def add_term(self, term: Term) -> Self:
        """Adds a glossary term to the entity.

        Args:
            term: The Term object representing the glossary term to be added.

        Returns:
            The patch builder instance.
        """
        # TODO: Make this support raw strings, in addition to Term objects.
        self._add_patch(
            GlossaryTermsClass.ASPECT_NAME, "add", path=("terms", term.urn), value=term
        )
        return self

    def remove_term(self, term: Union[str, Urn]) -> Self:
        """Removes a glossary term from the entity.

        Args:
            term: The term to remove, specified as a string or Urn object.

        Returns:
            The patch builder instance.
        """
        if isinstance(term, str) and not term.startswith("urn:li:glossaryTerm:"):
            term = GlossaryTermUrn(term)
        self._add_patch(
            GlossaryTermsClass.ASPECT_NAME, "remove", path=("terms", term), value={}
        )
        return self



import json
import logging
import traceback
import uuid
from typing import Any, Dict, List, Optional, Type, Union

from sqlalchemy import types
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.sql.visitors import Visitable

from datahub.ingestion.extractor.schema_util import avro_schema_to_mce_fields
from datahub.metadata.com.linkedin.pegasus2avro.schema import SchemaField
from datahub.metadata.schema_classes import NullTypeClass, SchemaFieldDataTypeClass

logger = logging.getLogger(__name__)

try:
    # This is used for both BigQuery and Athena.
    from sqlalchemy_bigquery import STRUCT
except ImportError:
    STRUCT = None


class MapType(types.TupleType):
    # Wrapper class around SQLalchemy's TupleType to increase compatibility with DataHub
    pass


class SqlAlchemyColumnToAvroConverter:
    """Helper class that collects some methods to convert SQLalchemy columns to Avro schema."""

    # tuple of complex data types that require a special handling
    _COMPLEX_TYPES = (STRUCT, types.ARRAY, MapType)

    # mapping of primitive SQLalchemy data types to AVRO schema data types
    PRIMITIVE_SQL_ALCHEMY_TYPE_TO_AVRO_TYPE: Dict[Type[types.TypeEngine], str] = {
        types.String: "string",
        types.BINARY: "string",
        types.BOOLEAN: "boolean",
        types.FLOAT: "float",
        types.INTEGER: "int",
        types.BIGINT: "long",
        types.VARCHAR: "string",
        types.CHAR: "string",
    }

    @classmethod
    def get_avro_type(
        cls, column_type: Union[types.TypeEngine, STRUCT, MapType], nullable: bool
    ) -> Dict[str, Any]:
        """Determines the concrete AVRO schema type for a SQLalchemy-typed column"""
        if isinstance(
            column_type, tuple(cls.PRIMITIVE_SQL_ALCHEMY_TYPE_TO_AVRO_TYPE.keys())
        ):
            return {
                "type": cls.PRIMITIVE_SQL_ALCHEMY_TYPE_TO_AVRO_TYPE[type(column_type)],
                "native_data_type": str(column_type),
                "_nullable": nullable,
            }
        if isinstance(column_type, types.DECIMAL):
            return {
                "type": "bytes",
                "logicalType": "decimal",
                "precision": int(column_type.precision),
                "scale": int(column_type.scale),
                "native_data_type": str(column_type),
                "_nullable": nullable,
            }
        if isinstance(column_type, types.DATE):
            return {
                "type": "int",
                "logicalType": "date",
                "native_data_type": str(column_type),
                "_nullable": nullable,
            }
        if isinstance(column_type, types.TIMESTAMP):
            return {
                "type": "long",
                "logicalType": "timestamp-millis",
                "native_data_type": str(column_type),
                "_nullable": nullable,
            }
        if isinstance(column_type, types.ARRAY):
            array_type = column_type.item_type

            return {
                "type": "array",
                "items": cls.get_avro_type(column_type=array_type, nullable=nullable),
                "native_data_type": f"array<{str(column_type.item_type)}>",
            }
        if isinstance(column_type, MapType):
            try:
                key_type = column_type.types[0]
                value_type = column_type.types[1]
                return {
                    "type": "map",
                    "values": cls.get_avro_type(
                        column_type=value_type, nullable=nullable
                    ),
                    "native_data_type": str(column_type),
                    "key_type": cls.get_avro_type(
                        column_type=key_type, nullable=nullable
                    ),
                    "key_native_data_type": str(key_type),
                }
            except Exception as e:
                logger.warning(
                    f"Unable to parse MapType {column_type} the error was: {e}"
                )
                return {
                    "type": "map",
                    "values": {"type": "null", "_nullable": True},
                    "native_data_type": str(column_type),
                    "key_type": {"type": "null", "_nullable": True},
                    "key_native_data_type": "null",
                }
        if STRUCT and isinstance(column_type, STRUCT):
            fields = []
            for field_def in column_type._STRUCT_fields:
                field_name, field_type = field_def
                fields.append(
                    {
                        "name": field_name,
                        "type": cls.get_avro_type(
                            column_type=field_type, nullable=nullable
                        ),
                    }
                )
            struct_name = f"__struct_{str(uuid.uuid4()).replace('-', '')}"
            try:
                return {
                    "type": "record",
                    "name": struct_name,
                    "fields": fields,
                    "native_data_type": str(column_type),
                    "_nullable": nullable,
                }
            except Exception:
                # This is a workaround for the case when the struct name is not string convertable because SqlAlchemt throws an error
                return {
                    "type": "record",
                    "name": struct_name,
                    "fields": fields,
                    "native_data_type": "map",
                    "_nullable": nullable,
                }

        return {
            "type": "null",
            "native_data_type": str(column_type),
            "_nullable": nullable,
        }

    @classmethod
    def get_avro_for_sqlalchemy_column(
        cls,
        column_name: str,
        column_type: types.TypeEngine,
        nullable: bool,
    ) -> Union[object, Dict[str, object]]:
        """Returns the AVRO schema representation of a SQLalchemy column."""
        if isinstance(column_type, cls._COMPLEX_TYPES):
            return {
                "type": "record",
                "name": "__struct_",
                "fields": [
                    {
                        "name": column_name,
                        "type": cls.get_avro_type(
                            column_type=column_type, nullable=nullable
                        ),
                    }
                ],
            }
        return cls.get_avro_type(column_type=column_type, nullable=nullable)


def get_schema_fields_for_sqlalchemy_column(
    column_name: str,
    column_type: types.TypeEngine,
    inspector: Inspector,
    description: Optional[str] = None,
    nullable: Optional[bool] = True,
    is_part_of_key: Optional[bool] = False,
    is_partitioning_key: Optional[bool] = False,
) -> List[SchemaField]:
    """Creates SchemaFields from a given SQLalchemy column.

    This function is analogous to `get_schema_fields_for_hive_column` from datahub.utilities.hive_schema_to_avro.
    The main purpose of implementing it this way, is to make it ready/compatible for second field path generation,
    which allows to explore nested structures within the UI.
    """

    if nullable is None:
        nullable = True

    try:
        # as a first step, the column is converted to AVRO JSON which can then be used by an existing function
        avro_schema_json = (
            SqlAlchemyColumnToAvroConverter.get_avro_for_sqlalchemy_column(
                column_name=column_name,
                column_type=column_type,
                nullable=nullable,
            )
        )
        # retrieve schema field definitions from the above generated AVRO JSON structure
        schema_fields = avro_schema_to_mce_fields(
            avro_schema=json.dumps(avro_schema_json),
            default_nullable=nullable,
            swallow_exceptions=False,
        )
    except Exception as e:
        logger.warning(
            f"Unable to parse column {column_name} and type {column_type} the error was: {e} Traceback: {traceback.format_exc()}"
        )

        # fallback description in case any exception occurred
        schema_fields = [
            SchemaField(
                fieldPath=column_name,
                type=SchemaFieldDataTypeClass(type=NullTypeClass()),
                nativeDataType=get_native_data_type_for_sqlalchemy_type(
                    column_type,
                    inspector,
                ),
            )
        ]

    # for all non-nested data types an additional modification of the `fieldPath` property is required
    if type(column_type) in (
        *SqlAlchemyColumnToAvroConverter.PRIMITIVE_SQL_ALCHEMY_TYPE_TO_AVRO_TYPE.keys(),
        types.TIMESTAMP,
        types.DATE,
        types.DECIMAL,
    ):
        schema_fields[0].fieldPath += f".{column_name}"

    if description:
        schema_fields[0].description = description
    schema_fields[0].isPartOfKey = (
        is_part_of_key if is_part_of_key is not None else False
    )

    schema_fields[0].isPartitioningKey = (
        is_partitioning_key if is_partitioning_key is not None else False
    )

    return schema_fields


def get_native_data_type_for_sqlalchemy_type(
    column_type: types.TypeEngine, inspector: Inspector
) -> str:
    if isinstance(column_type, types.NullType):
        return column_type.__visit_name__

    try:
        return column_type.compile(dialect=inspector.dialect)
    except Exception as e:
        logger.debug(
            f"Unable to compile sqlalchemy type {column_type} the error was: {e}"
        )

        if (
            isinstance(column_type, Visitable)
            and column_type.__visit_name__ is not None
        ):
            return column_type.__visit_name__

        return repr(column_type)



from collections import OrderedDict
from typing import Any, Tuple


def _pre_handle_union_with_aliases(
    obj: Any,
    from_pattern: str,
    to_pattern: str,
) -> Tuple[bool, Any]:
    # PDL supports "Unions with aliases", which are unions that have a field name
    # that is different from the type name.
    # See https://linkedin.github.io/rest.li/pdl_schema#union-with-aliases
    # When generating the avro schema, Rest.li adds a "fieldDiscriminator" field
    # to disambiguate between the different union types.

    if "fieldDiscriminator" in obj:
        # On the way out, we need to remove the field discriminator.
        field = obj["fieldDiscriminator"]
        return True, {
            field: _json_transform(obj[field], from_pattern, to_pattern, pre=True)
        }

    return False, None


def _post_handle_unions_with_aliases(
    obj: Any,
    from_pattern: str,
    to_pattern: str,
) -> Tuple[bool, Any]:
    # Note that "CostCostClass" is the only usage of a union with aliases in our
    # current PDL / metadata model, so the below hardcoding works.
    # Because this method is brittle and prone to becoming stale, we validate
    # the aforementioned assumption in our tests.

    if (
        set(obj.keys()) == {"cost", "costType"}
        and isinstance(obj["cost"], dict)
        and len(obj["cost"].keys()) == 1
        and list(obj["cost"].keys())[0] in {"costId", "costCode"}
    ):
        # On the way in, we need to add back the field discriminator.
        return True, {
            "cost": {
                **obj["cost"],
                "fieldDiscriminator": list(obj["cost"].keys())[0],
            },
            "costType": obj["costType"],
        }

    return False, None


def _json_transform(obj: Any, from_pattern: str, to_pattern: str, pre: bool) -> Any:
    if isinstance(obj, (dict, OrderedDict)):
        if len(obj.keys()) == 1:
            key: str = list(obj.keys())[0]
            value = obj[key]
            if key.startswith(from_pattern):
                new_key = key.replace(from_pattern, to_pattern, 1)
                return {
                    new_key: _json_transform(value, from_pattern, to_pattern, pre=pre)
                }

        if pre:
            handled, new_obj = _pre_handle_union_with_aliases(
                obj, from_pattern, to_pattern
            )
            if handled:
                return new_obj

        if not pre:
            handled, new_obj = _post_handle_unions_with_aliases(
                obj, from_pattern, to_pattern
            )
            if handled:
                return new_obj

        new_obj = {
            key: _json_transform(value, from_pattern, to_pattern, pre=pre)
            for key, value in obj.items()
            if value is not None
        }

        return new_obj
    elif isinstance(obj, list):
        new_obj = [
            _json_transform(item, from_pattern, to_pattern, pre=pre) for item in obj
        ]
        return new_obj
    elif isinstance(obj, bytes):
        return obj.decode()
    return obj


def pre_json_transform(obj: Any) -> Any:
    """Usually called before sending avro-serialized json over to the rest.li server"""
    return _json_transform(
        obj,
        from_pattern="com.linkedin.pegasus2avro.",
        to_pattern="com.linkedin.",
        pre=True,
    )


def post_json_transform(obj: Any) -> Any:
    """Usually called after receiving restli-serialized json before instantiating into avro-generated Python classes"""
    return _json_transform(
        obj,
        from_pattern="com.linkedin.",
        to_pattern="com.linkedin.pegasus2avro.",
        pre=False,
    )



import collections
import itertools
import json
import logging
import re
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import (
    Any,
    Callable,
    Dict,
    FrozenSet,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)
from urllib.parse import urlparse

import humanfriendly
import sqlglot
from google.cloud.datacatalog import lineage_v1
from google.cloud.logging_v2.client import Client as GCPLoggingClient

from datahub.api.entities.dataset.dataset import Dataset
from datahub.configuration.pattern_utils import is_schema_allowed
from datahub.emitter import mce_builder
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.source_helpers import auto_workunit
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.graph.client import DataHubGraph
from datahub.ingestion.source.bigquery_v2.bigquery_audit import (
    AuditLogEntry,
    BigQueryAuditMetadata,
    BigqueryTableIdentifier,
    BigQueryTableRef,
    QueryEvent,
    ReadEvent,
)
from datahub.ingestion.source.bigquery_v2.bigquery_audit_log_api import (
    BigQueryAuditLogApi,
)
from datahub.ingestion.source.bigquery_v2.bigquery_config import BigQueryV2Config
from datahub.ingestion.source.bigquery_v2.bigquery_report import BigQueryV2Report
from datahub.ingestion.source.bigquery_v2.bigquery_schema import (
    BigQuerySchemaApi,
    BigqueryTableSnapshot,
)
from datahub.ingestion.source.bigquery_v2.common import (
    BQ_DATETIME_FORMAT,
    BigQueryIdentifierBuilder,
)
from datahub.ingestion.source.bigquery_v2.queries import (
    BQ_FILTER_RULE_TEMPLATE_V2_LINEAGE,
    bigquery_audit_metadata_query_template_lineage,
)
from datahub.ingestion.source.gcs import gcs_utils
from datahub.ingestion.source.state.redundant_run_skip_handler import (
    RedundantLineageRunSkipHandler,
)
from datahub.ingestion.source_report.ingestion_stage import LINEAGE_EXTRACTION
from datahub.metadata.com.linkedin.pegasus2avro.schema import SchemaMetadata
from datahub.metadata.schema_classes import (
    AuditStampClass,
    DatasetLineageTypeClass,
    FineGrainedLineageClass,
    FineGrainedLineageDownstreamTypeClass,
    FineGrainedLineageUpstreamTypeClass,
    SchemaMetadataClass,
    UpstreamClass,
    UpstreamLineageClass,
)
from datahub.sql_parsing.schema_resolver import SchemaResolver
from datahub.sql_parsing.sql_parsing_aggregator import SqlParsingAggregator
from datahub.sql_parsing.sqlglot_lineage import SqlParsingResult, sqlglot_lineage
from datahub.utilities import memory_footprint
from datahub.utilities.file_backed_collections import FileBackedDict
from datahub.utilities.perf_timer import PerfTimer
from datahub.utilities.urns.dataset_urn import DatasetUrn

logger: logging.Logger = logging.getLogger(__name__)


@dataclass(order=True, eq=True, frozen=True)
class LineageEdgeColumnMapping:
    out_column: str
    in_columns: FrozenSet[str]


@dataclass(order=True, eq=True, frozen=True)
class LineageEdge:
    table: str  # in BigQueryTableRef format
    column_mapping: FrozenSet[LineageEdgeColumnMapping]

    auditStamp: datetime
    type: str = DatasetLineageTypeClass.TRANSFORMED
    column_confidence: float = 0.0


def _merge_lineage_edge_columns(
    a: Optional[LineageEdge], b: LineageEdge
) -> LineageEdge:
    if a is None:
        return b
    assert a.table == b.table

    merged_col_mapping: Dict[str, Set[str]] = collections.defaultdict(set)
    for col_mapping in itertools.chain(a.column_mapping, b.column_mapping):
        merged_col_mapping[col_mapping.out_column].update(col_mapping.in_columns)

    return LineageEdge(
        table=a.table,
        column_mapping=frozenset(
            LineageEdgeColumnMapping(
                out_column=out_column,
                in_columns=frozenset(in_columns),
            )
            for out_column, in_columns in merged_col_mapping.items()
        ),
        auditStamp=a.auditStamp,
        type=a.type,
        column_confidence=min(a.column_confidence, b.column_confidence),
    )


def _follow_column_lineage(
    temp: LineageEdge,
    upstream: LineageEdge,
) -> LineageEdge:
    """
    Collapse lineage of `base -> temp -> upstream` into `base -> upstream`.
    """

    # This is a mapping from temp's column -> upstream's columns.
    upstream_col_mapping: Dict[str, Set[str]] = collections.defaultdict(set)
    for col_mapping in upstream.column_mapping:
        upstream_col_mapping[col_mapping.out_column].update(col_mapping.in_columns)

    # This code can be a bit confusing.
    # - temp.column_mapping[i].out_column is a column in base.
    # - temp.column_mapping[i].in_columns are columns in temp.
    # - For each temp_col in temp.column_mapping[i].in_columns, we want to replace it with
    #   the columns in upstream that it depends on.
    return LineageEdge(
        table=upstream.table,
        column_mapping=frozenset(
            LineageEdgeColumnMapping(
                out_column=base_column_mapping.out_column,
                in_columns=frozenset(
                    col
                    for temp_col in base_column_mapping.in_columns
                    for col in upstream_col_mapping.get(temp_col, set())
                ),
            )
            for base_column_mapping in temp.column_mapping
        ),
        # We use the audit stamp and type from temp, since it's "closer" to the base.
        auditStamp=temp.auditStamp,
        type=temp.type,
        column_confidence=min(temp.column_confidence, upstream.column_confidence),
    )


def make_lineage_edges_from_parsing_result(
    sql_lineage: SqlParsingResult, audit_stamp: datetime, lineage_type: str
) -> List[LineageEdge]:
    # Note: This ignores the out_tables section of the sql parsing result.
    audit_stamp = datetime.now(timezone.utc)

    # Invariant: table_edges[k].table == k
    table_edges: Dict[str, LineageEdge] = {}

    # This isn't the most efficient code - it iterates over the same data multiple times,
    # once per in_table entry.
    for table_urn in sql_lineage.in_tables:
        column_mapping: Dict[str, FrozenSet[str]] = {}

        for column_lineage in sql_lineage.column_lineage or []:
            out_column = column_lineage.downstream.column
            column_mapping[out_column] = frozenset(
                upstream_column_info.column
                for upstream_column_info in column_lineage.upstreams
                if upstream_column_info.table == table_urn
            )

        try:
            table_name = str(
                BigQueryTableRef.from_bigquery_table(
                    BigqueryTableIdentifier.from_string_name(
                        DatasetUrn.from_string(table_urn).name
                    )
                )
            )
        except IndexError as e:
            logger.debug(f"Unable to parse table urn {table_urn}: {e}")
            continue

        table_edges[table_name] = LineageEdge(
            table=table_name,
            column_mapping=frozenset(
                LineageEdgeColumnMapping(out_column=out_column, in_columns=in_columns)
                for out_column, in_columns in column_mapping.items()
                if in_columns
            ),
            auditStamp=audit_stamp,
            type=lineage_type,
            column_confidence=sql_lineage.debug_info.confidence,
        )

    return list(table_edges.values())


class BigqueryLineageExtractor:
    def __init__(
        self,
        config: BigQueryV2Config,
        report: BigQueryV2Report,
        *,
        schema_resolver: SchemaResolver,
        identifiers: BigQueryIdentifierBuilder,
        redundant_run_skip_handler: Optional[RedundantLineageRunSkipHandler] = None,
    ):
        self.config = config
        self.report = report
        self.schema_resolver = schema_resolver

        self.identifiers = identifiers
        self.audit_log_api = BigQueryAuditLogApi(
            report.audit_log_api_perf,
            self.config.rate_limit,
            self.config.requests_per_min,
        )

        self.redundant_run_skip_handler = redundant_run_skip_handler
        self.start_time, self.end_time = (
            self.report.lineage_start_time,
            self.report.lineage_end_time,
        ) = self.get_time_window()

        self.datasets_skip_audit_log_lineage: Set[str] = set()

        self.aggregator = SqlParsingAggregator(
            platform=self.identifiers.platform,
            platform_instance=self.identifiers.identifier_config.platform_instance,
            env=self.identifiers.identifier_config.env,
            schema_resolver=self.schema_resolver,
            eager_graph_load=False,
            generate_lineage=True,
            generate_queries=True,
            generate_usage_statistics=False,
            generate_query_usage_statistics=False,
            generate_operations=False,
            format_queries=True,
        )
        self.report.sql_aggregator = self.aggregator.report
        self.gcs_uris_regex = re.compile(r"uris=\[([^\]]+)\]")

    def get_time_window(self) -> Tuple[datetime, datetime]:
        if self.redundant_run_skip_handler:
            return self.redundant_run_skip_handler.suggest_run_time_window(
                self.config.start_time, self.config.end_time
            )
        else:
            return self.config.start_time, self.config.end_time

    def _should_ingest_lineage(self) -> bool:
        if (
            self.redundant_run_skip_handler
            and self.redundant_run_skip_handler.should_skip_this_run(
                cur_start_time=self.config.start_time,
                cur_end_time=self.config.end_time,
            )
        ):
            # Skip this run
            self.report.warning(
                title="Skipped redundant lineage extraction",
                message="Skip this run as there was already a run for current ingestion window.",
            )
            return False

        return True

    def get_lineage_workunits_for_views_and_snapshots(
        self,
        projects: List[str],
        view_refs_by_project: Dict[str, Set[str]],
        view_definitions: FileBackedDict[str],
        snapshot_refs_by_project: Dict[str, Set[str]],
        snapshots_by_ref: FileBackedDict[BigqueryTableSnapshot],
    ) -> Iterable[MetadataWorkUnit]:
        for project in projects:
            for view in view_refs_by_project[project]:
                self.datasets_skip_audit_log_lineage.add(view)
                self.aggregator.add_view_definition(
                    view_urn=self.identifiers.gen_dataset_urn_from_raw_ref(
                        BigQueryTableRef.from_string_name(view)
                    ),
                    view_definition=view_definitions[view],
                    default_db=project,
                )

            for snapshot_ref in snapshot_refs_by_project[project]:
                snapshot = snapshots_by_ref[snapshot_ref]
                if not snapshot.base_table_identifier:
                    continue
                self.datasets_skip_audit_log_lineage.add(snapshot_ref)
                snapshot_urn = self.identifiers.gen_dataset_urn_from_raw_ref(
                    BigQueryTableRef.from_string_name(snapshot_ref)
                )
                base_table_urn = self.identifiers.gen_dataset_urn_from_raw_ref(
                    BigQueryTableRef(snapshot.base_table_identifier)
                )
                self.aggregator.add_known_lineage_mapping(
                    upstream_urn=base_table_urn, downstream_urn=snapshot_urn
                )

        yield from auto_workunit(self.aggregator.gen_metadata())

    def get_lineage_workunits(
        self,
        projects: List[str],
        table_refs: Set[str],
    ) -> Iterable[MetadataWorkUnit]:
        if not self._should_ingest_lineage():
            return

        if self.config.use_exported_bigquery_audit_metadata:
            projects = ["*"]  # project_id not used when using exported metadata

        for project in projects:
            with self.report.new_stage(f"{project}: {LINEAGE_EXTRACTION}"):
                yield from self.generate_lineage(
                    project,
                    table_refs,
                )

        if self.redundant_run_skip_handler:
            # Update the checkpoint state for this run.
            self.redundant_run_skip_handler.update_state(
                self.config.start_time, self.config.end_time
            )

    def generate_lineage(
        self,
        project_id: str,
        table_refs: Set[str],
    ) -> Iterable[MetadataWorkUnit]:
        logger.info(f"Generate lineage for {project_id}")
        with PerfTimer() as timer:
            try:
                if self.config.extract_lineage_from_catalog:
                    lineage = self.lineage_via_catalog_lineage_api(project_id)
                else:
                    events = self._get_parsed_audit_log_events(project_id)
                    lineage = self._create_lineage_map(events)
            except Exception as e:
                self.report.lineage_failed_extraction.append(project_id)
                self.report.warning(
                    title="Failed to extract lineage",
                    message="Unexpected error encountered",
                    context=project_id,
                    exc=e,
                )
                lineage = {}

            self.report.lineage_metadata_entries[project_id] = len(lineage)
            logger.info(f"Built lineage map containing {len(lineage)} entries.")
            logger.debug(f"lineage metadata is {lineage}")
            self.report.lineage_extraction_sec[project_id] = timer.elapsed_seconds(
                digits=2
            )
            self.report.lineage_mem_size[project_id] = humanfriendly.format_size(
                memory_footprint.total_size(lineage)
            )

        for lineage_key in lineage.keys():
            # For views, we do not use the upstreams obtained by parsing audit logs
            # as they may contain indirectly referenced tables.
            if (
                lineage_key not in table_refs
                or lineage_key in self.datasets_skip_audit_log_lineage
            ):
                continue

            yield from self.gen_lineage_workunits_for_table(
                lineage, BigQueryTableRef.from_string_name(lineage_key)
            )

    def gen_lineage_workunits_for_table(
        self, lineage: Dict[str, Set[LineageEdge]], table_ref: BigQueryTableRef
    ) -> Iterable[MetadataWorkUnit]:
        dataset_urn = self.identifiers.gen_dataset_urn_from_raw_ref(table_ref)

        lineage_info = self.get_lineage_for_table(
            bq_table=table_ref,
            bq_table_urn=dataset_urn,
            lineage_metadata=lineage,
        )
        if lineage_info:
            yield from self.gen_lineage(dataset_urn, lineage_info)

    def gen_lineage(
        self,
        dataset_urn: str,
        upstream_lineage: Optional[UpstreamLineageClass] = None,
    ) -> Iterable[MetadataWorkUnit]:
        if upstream_lineage is None:
            return

        if upstream_lineage is not None:
            if not self.config.extract_column_lineage:
                upstream_lineage.fineGrainedLineages = None

            # Incremental lineage is handled by the auto_incremental_lineage helper.
            yield from [
                MetadataChangeProposalWrapper(
                    entityUrn=dataset_urn, aspect=upstream_lineage
                ).as_workunit()
            ]

    def lineage_via_catalog_lineage_api(
        self, project_id: str
    ) -> Dict[str, Set[LineageEdge]]:
        """
        Uses Data Catalog API to request lineage metadata. Please take a look at the API documentation for more details.

        NOTE: It's necessary for you to enable the service API in your Google Cloud Project.

        Args:
            project_id(str): Google project id. Used to search for tables and datasets.

        Returns:
            Dict[str, Set[str]] - A dictionary, where keys are the downstream table's identifier and values is a set
            of upstream tables identifiers.
        """
        logger.info("Populating lineage info via Catalog Data Linage API")

        # Regions to search for BigQuery tables: projects/{project_id}/locations/{region}
        enabled_regions: List[str] = ["US", "EU"]

        lineage_client: lineage_v1.LineageClient = lineage_v1.LineageClient()

        data_dictionary = BigQuerySchemaApi(
            self.report.schema_api_perf,
            self.config.get_bigquery_client(),
            self.config.get_projects_client(),
        )

        # Filtering datasets
        datasets = list(data_dictionary.get_datasets_for_project_id(project_id))
        project_tables = []
        for dataset in datasets:
            # Enables only tables where type is TABLE, VIEW or MATERIALIZED_VIEW (not EXTERNAL)
            project_tables.extend(
                [
                    table
                    for table in data_dictionary.list_tables(dataset.name, project_id)
                    if table.table_type in ["TABLE", "VIEW", "MATERIALIZED_VIEW"]
                ]
            )

        lineage_map: Dict[str, Set[LineageEdge]] = {}
        curr_date = datetime.now()
        for project_table in project_tables:
            # Convert project table to <project_id>.<dataset_id>.<table_id> format
            table = f"{project_table.project}.{project_table.dataset_id}.{project_table.table_id}"

            if not is_schema_allowed(
                self.config.dataset_pattern,
                schema_name=project_table.dataset_id,
                db_name=project_table.project,
                match_fully_qualified_schema_name=self.config.match_fully_qualified_names,
            ) or not self.config.table_pattern.allowed(table):
                self.report.num_skipped_lineage_entries_not_allowed[
                    project_table.project
                ] += 1
                continue

            logger.info("Creating lineage map for table %s", table)
            upstreams = set()
            downstream_table = lineage_v1.EntityReference()
            # fully_qualified_name in format: "bigquery:<project_id>.<dataset_id>.<table_id>"
            downstream_table.fully_qualified_name = f"bigquery:{table}"
            # Searches in different regions
            for region in enabled_regions:
                location_request = lineage_v1.SearchLinksRequest(
                    target=downstream_table,
                    parent=f"projects/{project_id}/locations/{region.lower()}",
                )
                response = lineage_client.search_links(request=location_request)
                upstreams.update(
                    [
                        str(lineage.source.fully_qualified_name).replace(
                            "bigquery:", ""
                        )
                        for lineage in response
                    ]
                )

            # Downstream table identifier
            destination_table_str = str(
                BigQueryTableRef(
                    table_identifier=BigqueryTableIdentifier(*table.split("."))
                )
            )

            # Only builds lineage map when the table has upstreams
            logger.debug("Found %d upstreams for table %s", len(upstreams), table)
            if upstreams:
                lineage_map[destination_table_str] = {
                    LineageEdge(
                        table=str(
                            BigQueryTableRef(
                                table_identifier=BigqueryTableIdentifier.from_string_name(
                                    source_table
                                )
                            )
                        ),
                        column_mapping=frozenset(),
                        auditStamp=curr_date,
                    )
                    for source_table in upstreams
                }
        return lineage_map

    def _get_parsed_audit_log_events(self, project_id: str) -> Iterable[QueryEvent]:
        # We adjust the filter values a bit, since we need to make sure that the join
        # between query events and read events is complete. For example, this helps us
        # handle the case where the read happens within our time range but the query
        # completion event is delayed and happens after the configured end time.
        corrected_start_time = self.start_time - self.config.max_query_duration
        corrected_end_time = self.end_time + self.config.max_query_duration
        self.report.log_entry_start_time = corrected_start_time
        self.report.log_entry_end_time = corrected_end_time

        parse_fn: Callable[[Any], Optional[Union[ReadEvent, QueryEvent]]]
        if self.config.use_exported_bigquery_audit_metadata:
            entries = self.get_exported_log_entries(
                corrected_start_time, corrected_end_time
            )
            parse_fn = self._parse_exported_bigquery_audit_metadata
        else:
            entries = self.get_log_entries_via_gcp_logging(
                project_id, corrected_start_time, corrected_end_time
            )
            parse_fn = self._parse_bigquery_log_entries

        for entry in entries:
            self.report.num_lineage_total_log_entries[project_id] += 1
            try:
                event = parse_fn(entry)
                if event:
                    self.report.num_lineage_parsed_log_entries[project_id] += 1
                    yield event
            except Exception as e:
                logger.warning(f"Unable to parse log entry `{entry}`: {e}")
                self.report.num_lineage_log_parse_failures[project_id] += 1

    def get_exported_log_entries(
        self, corrected_start_time, corrected_end_time, limit=None
    ):
        logger.info("Populating lineage info via exported GCP audit logs")
        bq_client = self.config.get_bigquery_client()
        entries = self.audit_log_api.get_exported_bigquery_audit_metadata(
            bigquery_client=bq_client,
            bigquery_audit_metadata_query_template=bigquery_audit_metadata_query_template_lineage,
            bigquery_audit_metadata_datasets=self.config.bigquery_audit_metadata_datasets,
            use_date_sharded_audit_log_tables=self.config.use_date_sharded_audit_log_tables,
            start_time=corrected_start_time,
            end_time=corrected_end_time,
            limit=limit,
        )
        return entries

    def get_log_entries_via_gcp_logging(
        self, project_id, corrected_start_time, corrected_end_time
    ):
        logger.info("Populating lineage info via exported GCP audit logs")

        logging_client = self.config.make_gcp_logging_client(project_id)
        logger.info(
            f"Start loading log entries from BigQuery for {project_id} "
            f"with start_time={corrected_start_time} and end_time={corrected_end_time}"
        )
        entries = self.audit_log_api.get_bigquery_log_entries_via_gcp_logging(
            logging_client,
            BQ_FILTER_RULE_TEMPLATE_V2_LINEAGE.format(
                start_time=corrected_start_time.strftime(BQ_DATETIME_FORMAT),
                end_time=corrected_end_time.strftime(BQ_DATETIME_FORMAT),
            ),
            self.config.log_page_size,
        )
        return entries

    # Currently we only parse JobCompleted events but in future we would want to parse other
    # events to also create field level lineage.
    def _parse_bigquery_log_entries(
        self,
        entry: AuditLogEntry,
    ) -> Optional[QueryEvent]:
        event: Optional[QueryEvent] = None

        missing_entry = QueryEvent.get_missing_key_entry(entry=entry)
        if missing_entry is None:
            event = QueryEvent.from_entry(
                entry,
                debug_include_full_payloads=self.config.debug_include_full_payloads,
            )

        missing_entry_v2 = QueryEvent.get_missing_key_entry_v2(entry=entry)
        if event is None and missing_entry_v2 is None:
            event = QueryEvent.from_entry_v2(
                entry, self.config.debug_include_full_payloads
            )

        if event is None:
            logger.warning(
                f"Unable to parse log missing {missing_entry}, missing v2 {missing_entry_v2} for {entry}",
            )
            return None
        else:
            return event

    def _parse_exported_bigquery_audit_metadata(
        self, audit_metadata: BigQueryAuditMetadata
    ) -> Optional[QueryEvent]:
        event: Optional[QueryEvent] = None

        missing_exported_audit = (
            QueryEvent.get_missing_key_exported_bigquery_audit_metadata(audit_metadata)
        )

        if missing_exported_audit is None:
            event = QueryEvent.from_exported_bigquery_audit_metadata(
                audit_metadata, self.config.debug_include_full_payloads
            )

        if event is None:
            logger.warning(
                f"Unable to parse audit metadata missing {missing_exported_audit} for {audit_metadata}",
            )
            return None
        else:
            return event

    def _create_lineage_map(
        self,
        entries: Iterable[QueryEvent],
    ) -> Dict[str, Set[LineageEdge]]:
        logger.info("Entering create lineage map function")
        lineage_map: Dict[str, Set[LineageEdge]] = collections.defaultdict(set)
        for e in entries:
            self.report.num_total_lineage_entries[e.project_id] = (
                self.report.num_total_lineage_entries.get(e.project_id, 0) + 1
            )

            ts = e.end_time if e.end_time else datetime.now(tz=timezone.utc)

            # Note that we trust the audit log to have the correct destination table.
            destination_table = e.destinationTable
            destination_table_str = str(destination_table)
            if not destination_table:
                self.report.num_skipped_lineage_entries_missing_data[e.project_id] += 1
                continue

            if not is_schema_allowed(
                self.config.dataset_pattern,
                destination_table.table_identifier.dataset,
                destination_table.table_identifier.project_id,
                self.config.match_fully_qualified_names,
            ) or not self.config.table_pattern.allowed(
                destination_table.table_identifier.get_table_name()
            ):
                self.report.num_skipped_lineage_entries_not_allowed[e.project_id] += 1
                continue

            # If there is a view being referenced then bigquery sends both the view as well as underlying table
            # in the references. There is no distinction between direct/base objects accessed. Doing sql parsing
            # ensures that we only use direct objects accessed for lineage.
            #
            # Because of this, we use the SQL parser as the primary source of truth for lineage
            # and the audit log as a fallback.
            lineage_from_event: Set[LineageEdge] = set()

            # Try the sql parser first.
            if self.config.lineage_use_sql_parser:
                logger.debug(
                    f"Using sql parser for lineage extraction for destination table: {destination_table.table_identifier.get_table_name()}, queryType: {e.statementType}, query: {e.query}"
                )
                if e.statementType == "SELECT":
                    # We wrap select statements in a CTE to make them parseable as insert statement.
                    # This is a workaround for the sql parser to support the case where the user runs a query and inserts the result into a table..
                    try:
                        parsed_queries = sqlglot.parse(e.query, "bigquery")
                        if parsed_queries[-1]:
                            query = f"""create table `{destination_table.get_sanitized_table_ref().table_identifier.get_table_name()}` AS
                            (
                                {parsed_queries[-1].sql(dialect="bigquery")}
                            )"""
                        else:
                            query = e.query
                    except Exception:
                        logger.debug(
                            f"Failed to parse select-based lineage query {e.query} for table {destination_table}."
                            "Sql parsing will likely fail for this query, which will result in a fallback to audit log."
                        )
                        query = e.query
                else:
                    query = e.query
                raw_lineage = sqlglot_lineage(
                    query,
                    schema_resolver=self.schema_resolver,
                    default_db=e.project_id,
                )
                logger.debug(
                    f"Input tables: {raw_lineage.in_tables}, Output tables: {raw_lineage.out_tables}"
                )
                if raw_lineage.debug_info.table_error:
                    logger.debug(
                        f"Sql Parser failed on query: {e.query}. It won't cause any major issues, but "
                        f"queries referencing views may contain extra lineage for the tables underlying those views. "
                        f"The error was {raw_lineage.debug_info.table_error}."
                    )
                    self.report.num_lineage_entries_sql_parser_failure[
                        e.project_id
                    ] += 1

                lineage_from_event = set(
                    make_lineage_edges_from_parsing_result(
                        raw_lineage,
                        audit_stamp=ts,
                        lineage_type=DatasetLineageTypeClass.TRANSFORMED,
                    )
                )

            if not lineage_from_event:
                # Fallback to the audit log if the sql parser didn't find anything.
                if not (e.referencedTables or e.referencedViews):
                    self.report.num_skipped_lineage_entries_missing_data[
                        e.project_id
                    ] += 1
                    continue

                for ref_table_or_view in [*e.referencedTables, *e.referencedViews]:
                    if str(ref_table_or_view) != destination_table_str:
                        lineage_from_event.add(
                            LineageEdge(
                                table=str(ref_table_or_view),
                                auditStamp=ts,
                                column_mapping=frozenset(),
                                column_confidence=0.1,
                            )
                        )

            if not lineage_from_event:
                self.report.num_skipped_lineage_entries_other[e.project_id] += 1
                continue

            lineage_map[destination_table_str].update(lineage_from_event)

        logger.info("Exiting create lineage map function")
        return lineage_map

    def get_upstream_tables(
        self,
        bq_table: BigQueryTableRef,
        lineage_metadata: Dict[str, Set[LineageEdge]],
        edges_seen: Optional[Set[LineageEdge]] = None,
    ) -> Set[LineageEdge]:
        if edges_seen is None:
            edges_seen = set()

        upstreams: Dict[str, LineageEdge] = {}
        for upstream_lineage in lineage_metadata[str(bq_table)]:
            upstream_table_ref = upstream_lineage.table
            upstream_table = BigQueryTableRef.from_string_name(upstream_table_ref)
            if bq_table == upstream_table:
                # Skip self-references.
                continue

            if upstream_table.is_temporary_table(
                [self.config.temp_table_dataset_prefix]
            ):
                # making sure we don't process a table twice and not get into a recursive loop
                if upstream_lineage in edges_seen:
                    logger.debug(
                        f"Skipping table {upstream_lineage} because it was seen already"
                    )
                    continue
                edges_seen.add(upstream_lineage)

                if upstream_table_ref in lineage_metadata:
                    # `upstream_table` is a temporary table.
                    # We don't want it in the lineage, but we do want its upstreams.
                    # When following lineage for a temp table, we need to merge the column lineage.

                    for temp_table_upstream in self.get_upstream_tables(
                        upstream_table,
                        lineage_metadata=lineage_metadata,
                        edges_seen=edges_seen,
                    ):
                        ref_temp_table_upstream = temp_table_upstream.table

                        # Replace `bq_table -> upstream_table -> temp_table_upstream`
                        # with `bq_table -> temp_table_upstream`, merging the column lineage.
                        collapsed_lineage = _follow_column_lineage(
                            upstream_lineage, temp_table_upstream
                        )

                        upstreams[ref_temp_table_upstream] = (
                            _merge_lineage_edge_columns(
                                upstreams.get(ref_temp_table_upstream),
                                collapsed_lineage,
                            )
                        )
            else:
                upstreams[upstream_table_ref] = _merge_lineage_edge_columns(
                    upstreams.get(upstream_table_ref),
                    upstream_lineage,
                )

        return set(upstreams.values())

    def get_lineage_for_table(
        self,
        bq_table: BigQueryTableRef,
        bq_table_urn: str,
        lineage_metadata: Dict[str, Set[LineageEdge]],
    ) -> Optional[UpstreamLineageClass]:
        upstream_list: List[UpstreamClass] = []
        fine_grained_lineages: List[FineGrainedLineageClass] = []
        # Sorting the list of upstream lineage events in order to avoid creating multiple aspects in backend
        # even if the lineage is same but the order is different.
        for upstream in sorted(self.get_upstream_tables(bq_table, lineage_metadata)):
            upstream_table = BigQueryTableRef.from_string_name(upstream.table)
            upstream_table_urn = self.identifiers.gen_dataset_urn_from_raw_ref(
                upstream_table
            )

            # Generate table-level lineage.
            upstream_table_class = UpstreamClass(
                dataset=upstream_table_urn,
                type=upstream.type,
                auditStamp=AuditStampClass(
                    actor="urn:li:corpuser:datahub",
                    time=int(upstream.auditStamp.timestamp() * 1000),
                ),
            )
            if self.config.upstream_lineage_in_report:
                current_lineage_map: Set = self.report.upstream_lineage.get(
                    str(bq_table), set()
                )
                current_lineage_map.add(str(upstream_table))
                self.report.upstream_lineage[str(bq_table)] = current_lineage_map

            upstream_list.append(upstream_table_class)

            # Generate column-level lineage.
            for col_lineage_edge in upstream.column_mapping:
                fine_grained_lineage = FineGrainedLineageClass(
                    downstreamType=FineGrainedLineageDownstreamTypeClass.FIELD,
                    downstreams=[
                        mce_builder.make_schema_field_urn(
                            bq_table_urn, col_lineage_edge.out_column
                        )
                    ],
                    upstreamType=FineGrainedLineageUpstreamTypeClass.FIELD_SET,
                    upstreams=[
                        mce_builder.make_schema_field_urn(
                            upstream_table_urn, upstream_col
                        )
                        for upstream_col in col_lineage_edge.in_columns
                    ],
                    confidenceScore=upstream.column_confidence,
                )
                fine_grained_lineages.append(fine_grained_lineage)

        if upstream_list:
            upstream_lineage = UpstreamLineageClass(
                upstreams=upstream_list,
                fineGrainedLineages=fine_grained_lineages,
            )
            return upstream_lineage

        return None

    def test_capability(self, project_id: str) -> None:
        if self.config.use_exported_bigquery_audit_metadata:
            for entry in self.get_exported_log_entries(
                self.start_time,
                self.end_time,
                limit=1,
            ):
                logger.debug(
                    f"Connection test got one exported_bigquery_audit_metadata {entry}"
                )
        else:
            gcp_logging_client: GCPLoggingClient = self.config.make_gcp_logging_client(
                project_id
            )
            for entry in self.audit_log_api.get_bigquery_log_entries_via_gcp_logging(
                gcp_logging_client,
                filter=BQ_FILTER_RULE_TEMPLATE_V2_LINEAGE.format(
                    start_time=self.start_time.strftime(BQ_DATETIME_FORMAT),
                    end_time=self.end_time.strftime(BQ_DATETIME_FORMAT),
                ),
                log_page_size=self.config.log_page_size,
                limit=1,
            ):
                logger.debug(f"Connection test got one audit metadata entry {entry}")

    def report_status(self, step: str, status: bool) -> None:
        if self.redundant_run_skip_handler:
            self.redundant_run_skip_handler.report_current_run_status(step, status)

    def gen_lineage_workunits_for_external_table(
        self,
        dataset_urn: str,
        ddl: Optional[str],
        graph: Optional[DataHubGraph] = None,
    ) -> Iterable[MetadataWorkUnit]:
        if not ddl:
            return

        # Expect URIs in `uris=[""]` format
        uris_match = self.gcs_uris_regex.search(ddl)
        if not uris_match:
            self.report.num_skipped_external_table_lineage += 1
            logger.warning(f"Unable to parse GCS URI from the provided DDL {ddl}.")
            return

        uris_str = uris_match.group(1)
        try:
            source_uris = json.loads(f"[{uris_str}]")
        except json.JSONDecodeError as e:
            self.report.num_skipped_external_table_lineage += 1
            logger.warning(
                f"Json load failed on loading source uri with error: {e}. The field value was: {uris_str}"
            )
            return

        lineage_info = self.get_lineage_for_external_table(
            dataset_urn=dataset_urn,
            source_uris=source_uris,
            graph=graph,
        )

        if lineage_info:
            yield MetadataChangeProposalWrapper(
                entityUrn=dataset_urn, aspect=lineage_info
            ).as_workunit()

    def get_lineage_for_external_table(
        self,
        dataset_urn: str,
        source_uris: List[str],
        graph: Optional[DataHubGraph] = None,
    ) -> Optional[UpstreamLineageClass]:
        upstreams_list: List[UpstreamClass] = []
        fine_grained_lineages: List[FineGrainedLineageClass] = []
        gcs_urns: Set[str] = set()

        for source_uri in source_uris:
            # Check that storage_location have the gs:// prefix.
            # Right now we are only supporting GCS lineage
            if not gcs_utils.is_gcs_uri(source_uri):
                continue
            gcs_path = self._get_gcs_path(source_uri)

            if gcs_path is None:
                continue

            path = gcs_utils.strip_gcs_prefix(gcs_path)
            urn = mce_builder.make_dataset_urn_with_platform_instance(
                platform="gcs",
                name=path,
                env=self.config.env,
                platform_instance=(
                    self.config.platform_instance
                    if self.config.platform_instance is not None
                    else None
                ),
            )
            gcs_urns.add(urn)

        upstreams_list.extend(
            [
                UpstreamClass(
                    dataset=source_dataset_urn,
                    type=DatasetLineageTypeClass.COPY,
                )
                for source_dataset_urn in gcs_urns
            ]
        )

        if not upstreams_list:
            return None

        if self.config.include_column_lineage_with_gcs:
            assert graph
            schema_metadata: Optional[SchemaMetadataClass] = graph.get_schema_metadata(
                dataset_urn
            )
            for gcs_dataset_urn in gcs_urns:
                schema_metadata_for_gcs: Optional[SchemaMetadataClass] = (
                    graph.get_schema_metadata(gcs_dataset_urn)
                )
                if schema_metadata and schema_metadata_for_gcs:
                    fine_grained_lineage = self.get_fine_grained_lineages_with_gcs(
                        dataset_urn,
                        gcs_dataset_urn,
                        schema_metadata,
                        schema_metadata_for_gcs,
                    )
                    if not fine_grained_lineage:
                        logger.warning(
                            f"Failed to retrieve fine-grained lineage for dataset {dataset_urn} and GCS {gcs_dataset_urn}. "
                            f"Check schema metadata: {schema_metadata} and GCS metadata: {schema_metadata_for_gcs}."
                        )
                        continue

                    fine_grained_lineages.extend(fine_grained_lineage)

        upstream_lineage = UpstreamLineageClass(
            upstreams=upstreams_list, fineGrainedLineages=fine_grained_lineages or None
        )
        return upstream_lineage

    def _get_gcs_path(self, path: str) -> Optional[str]:
        if self.config.gcs_lineage_config:
            for path_spec in self.config.gcs_lineage_config.path_specs:
                if not path_spec.allowed(path):
                    logger.debug(
                        f"Skipping gcs path {path} as it does not match any path spec."
                    )
                    self.report.num_lineage_dropped_gcs_path += 1
                    continue

                _, table_path = path_spec.extract_table_name_and_path(path)
                return table_path

            if (
                self.config.gcs_lineage_config.ignore_non_path_spec_path
                and len(self.config.gcs_lineage_config.path_specs) > 0
            ):
                self.report.num_lineage_dropped_gcs_path += 1
                logger.debug(
                    f"Skipping gcs path {path} as it does not match any path spec."
                )
                return None

            if self.config.gcs_lineage_config.strip_urls:
                if "/" in urlparse(path).path:
                    return str(path.rsplit("/", 1)[0])

        return path

    def get_fine_grained_lineages_with_gcs(
        self,
        dataset_urn: str,
        gcs_dataset_urn: str,
        schema_metadata: SchemaMetadata,
        schema_metadata_for_gcs: SchemaMetadata,
    ) -> Optional[List[FineGrainedLineageClass]]:
        def simplify_field_path(field_path):
            return Dataset._simplify_field_path(field_path)

        if schema_metadata and schema_metadata_for_gcs:
            fine_grained_lineages: List[FineGrainedLineageClass] = []
            for field in schema_metadata.fields:
                field_path_v1 = simplify_field_path(field.fieldPath)
                matching_gcs_field = next(
                    (
                        f
                        for f in schema_metadata_for_gcs.fields
                        if simplify_field_path(f.fieldPath) == field_path_v1
                    ),
                    None,
                )
                if matching_gcs_field:
                    fine_grained_lineages.append(
                        FineGrainedLineageClass(
                            downstreamType=FineGrainedLineageDownstreamTypeClass.FIELD,
                            downstreams=[
                                mce_builder.make_schema_field_urn(
                                    dataset_urn, field_path_v1
                                )
                            ],
                            upstreamType=FineGrainedLineageUpstreamTypeClass.FIELD_SET,
                            upstreams=[
                                mce_builder.make_schema_field_urn(
                                    gcs_dataset_urn,
                                    simplify_field_path(matching_gcs_field.fieldPath),
                                )
                            ],
                        )
                    )
            return fine_grained_lineages
        return None



import collections
import threading
import time
from contextlib import AbstractContextManager
from typing import Any, Deque


# Modified version of https://github.com/RazerM/ratelimiter/blob/master/ratelimiter/_sync.py
class RateLimiter(AbstractContextManager):
    """Provides rate limiting for an operation with a configurable number of
    requests for a time period.
    """

    def __init__(self, max_calls: int, period: float = 1.0) -> None:
        """Initialize a RateLimiter object which enforces as much as max_calls
        operations on period (eventually floating) number of seconds.
        """
        if period <= 0:
            raise ValueError("Rate limiting period should be > 0")
        if max_calls <= 0:
            raise ValueError("Rate limiting number of calls should be > 0")

        # We're using a deque to store the last execution timestamps, not for
        # its maxlen attribute, but to allow constant time front removal.
        self.calls: Deque = collections.deque()

        self.period = period
        self.max_calls = max_calls
        self._lock = threading.Lock()

    def __enter__(self) -> "RateLimiter":
        with self._lock:
            # We want to ensure that no more than max_calls were run in the allowed
            # period. For this, we store the last timestamps of each call and run
            # the rate verification upon each __enter__ call.
            if len(self.calls) >= self.max_calls:
                until = time.time() + self.period - self._timespan
                sleeptime = until - time.time()
                if sleeptime > 0:
                    time.sleep(sleeptime)
            return self

    def __exit__(self, exc_type: Any, exc: Any, traceback: Any) -> None:
        with self._lock:
            # Store the last operation timestamp.
            self.calls.append(time.time())

            # Pop the timestamp list front (ie: the older calls) until the sum goes
            # back below the period. This is our 'sliding period' window.
            while self._timespan >= self.period:
                self.calls.popleft()

    @property
    def _timespan(self) -> float:
        return self.calls[-1] - self.calls[0]



from typing import Dict, List, Optional

from datahub.configuration.common import ConfigModel


class DatahubClientConfig(ConfigModel):
    """Configuration class for holding connectivity to datahub gms"""

    # TODO: Having a default for the server doesn't make a ton of sense. This should be handled
    # by callers / the CLI, but the actual client should not have any magic.
    server: str
    token: Optional[str] = None
    timeout_sec: Optional[float] = None
    retry_status_codes: Optional[List[int]] = None
    retry_max_times: Optional[int] = None
    extra_headers: Optional[Dict[str, str]] = None
    ca_certificate_path: Optional[str] = None
    client_certificate_path: Optional[str] = None
    disable_ssl_verification: bool = False



from typing import Callable, Iterable, Optional, Union

from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.common import ControlRecord, PipelineContext, RecordEnvelope
from datahub.ingestion.api.transform import Transformer
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.metadata.com.linkedin.pegasus2avro.mxe import (
    MetadataChangeEvent,
    MetadataChangeProposal,
)


class AutoHelperTransformer(Transformer):
    """Converts an auto_* source helper into a transformer.

    Important usage note: this assumes that the auto helper is stateless. The converter
    will be called multiple times, once for each batch of records. If the helper
    attempts to maintain state or perform some cleanup at the end of the stream, it
    will not behave correctly.
    """

    def __init__(
        self,
        converter: Callable[[Iterable[MetadataWorkUnit]], Iterable[MetadataWorkUnit]],
    ):
        self.converter = converter

    def transform(
        self, record_envelopes: Iterable[RecordEnvelope]
    ) -> Iterable[RecordEnvelope]:
        records = list(record_envelopes)

        normal_records = [r for r in records if not isinstance(r.record, ControlRecord)]
        control_records = [r for r in records if isinstance(r.record, ControlRecord)]

        yield from self._from_workunits(
            self.converter(
                self._into_workunits(normal_records),
            )
        )

        # Pass through control records as-is. Note that this isn't fully correct, since it technically
        # reorders the control records relative to the normal records. This is ok since the only control
        # record we have marks the end of the stream.
        yield from control_records

    @classmethod
    def _into_workunits(
        cls,
        stream: Iterable[
            RecordEnvelope[
                Union[
                    MetadataChangeEvent,
                    MetadataChangeProposal,
                    MetadataChangeProposalWrapper,
                ]
            ]
        ],
    ) -> Iterable[MetadataWorkUnit]:
        for record in stream:
            workunit_id: Optional[str] = record.metadata.get("workunit_id")
            metadata = record.record
            yield MetadataWorkUnit.from_metadata(metadata, id=workunit_id)

    @classmethod
    def _from_workunits(
        cls, stream: Iterable[MetadataWorkUnit]
    ) -> Iterable[RecordEnvelope]:
        for workunit in stream:
            yield RecordEnvelope(
                workunit.metadata,
                {
                    "workunit_id": workunit.id,
                },
            )

    @classmethod
    def create(cls, config_dict: dict, ctx: PipelineContext) -> Transformer:
        raise NotImplementedError(f"{cls.__name__} cannot be created from config")
